use reqwest;

// use std::io::Write;
// use std::io::Read;

use scraper;
use std::{
    collections::HashMap,
    fs::File,
    io::Write,
    path::{Path, PathBuf},
    string::String,
    sync::LazyLock,
};

use clap::{Parser, Subcommand};
use std::{thread, time::Duration};

#[derive(Parser)]
#[command(about, long_about=None)]
struct Cli {
    /// Working directory
    #[arg(short, long, default_value = "bible")]
    work: PathBuf,

    /// Bible Version
    #[arg(short, long, default_value = "ESV")]
    bible: String,

    #[command(subcommand)]
    command: Option<Commands>,
}

#[derive(Subcommand)]
enum Commands {
    /// Fetch from the web, using work directory for cache
    Fetch {
        /// Delay
        #[arg(short, long, default_value = "10")]
        delay: u32,
    },
    /// Extract information from cached files
    Extract {},
    /// Test something out
    Test {},
}

static APP_USER_AGENT: &str =
    "Mozilla/5.0 (X11; Linux x86_64; rv:134.0) Gecko/20100101 Firefox/134.0";

static BASE_URL: &str = "https://www.bible.com";

static VERSION_URLS: LazyLock<HashMap<&str, &str>> = LazyLock::new(|| {
    HashMap::from([
        ("ESV", "https://www.bible.com/bible/59/GEN.1.ESV"),
        ("KJV", "https://www.bible.com/bible/1/GEN.1.KJV"),
        ("NIV", "https://www.bible.com/bible/111/GEN.INTRO1.NIV"),
        // https://www.bible.com/bible/111/GEN.1.NIV"),
        ("YLT98", "https://www.bible.com/bible/821/GEN.1.YLT98"),
    ])
});

static BOOKS: LazyLock<Vec<&str>> = LazyLock::new(|| {
    Vec::from([
        "GEN", "EXO", "LEV", "NUM", "DEU", "JOS", "JDG", "RUT", "1SA", "2SA", "1KI", "2KI", "1CH",
        "2CH", "EZR", "NEH", "EST", "JOB", "PSA", "PRO", "ECC", "SNG", "ISA", "JER", "LAM", "EZK",
        "DAN", "HOS", "JOL", "AMO", "OBA", "JON", "MIC", "NAM", "HAB", "ZEP", "HAG", "ZEC", "MAL",
        "MAT", "MRK", "LUK", "JHN", "ACT", "ROM", "1CO", "2CO", "GAL", "EPH", "PHP", "COL", "1TH",
        "2TH", "1TI", "2TI", "TIT", "PHM", "HEB", "JAS", "1PE", "2PE", "1JN", "2JN", "3JN", "JUD",
        "REV",
    ])
});

// find_files in base_dir that end with extension bible.
fn find_files(base_dir: &str, bible: &str) -> Vec<String> {
    let paths = std::fs::read_dir(base_dir).unwrap();
    let mut result = Vec::<String>::new();

    for path in paths {
        if let Ok(dir) = path {
            let filename = dir.file_name().to_string_lossy().to_string();
            if filename.ends_with(bible) {
                // result.push(dir.file_name().to_string_lossy().to_string());
                result.push(dir.path().to_string_lossy().to_string());
            }
        }
    }
    result
}

// Start here
// static URL: &str = "https://www.bible.com/bible/59/PSA.95.ESV";
// "https://www.bible.com/bible/59/GEN.1.ESV";

// And maybe:
// https://www.bible.com/bible/2692/GEN.1.NASB2020
// https://www.bible.com/bible/1/GEN.1.KJV
// https://www.bible.com/bible/2692/GEN.1.NASB2020
// https://www.bible.com/bible/111/GEN.1.NIV
// https://www.bible.com/bible/821/GEN.1.YLT98
// Catholic
// https://www.bible.com/bible/42/GEN.1.CPDV

// Audio
// https://www.bible.com/audio-bible/59/GEN.1.ESV
// <script type="application/ld+json">{"@context":"https://schema.org","@type":"AudioObject","mainEntityOfPage":{"@type":"WebPage","@id":"https://www.bible.com/audio-bible/59/GEN.1.ESV"},"headline":"Audio Bible: Listen to Genesis 1 English Standard Version 2016 ESV","contentUrl":"https://audio-bible-cdn.youversionapi.com/1/32k/GEN/1-9dcefc68c6f7244489f59475fc7a1866.mp3?version_id=59",

// https://www.bible.com/verse-of-the-day

struct FetchResult {
    cached: bool,
    html: String,
}

fn fetch_cache(client: &reqwest::blocking::Client, url: &str) -> FetchResult {
    let (_, filename) = url.rsplit_once('/').unwrap();
    let path = Path::new("bible").join(filename);

    if path.exists() {
        // File already exists -- use cached version.
        let buffer = std::fs::read_to_string(path).unwrap();
        return FetchResult {
            cached: true,
            html: buffer,
        };
    }

    println!("fetch_cache {} => {}", url, filename);
    let res = client.get(url).send().unwrap();
    let buffer = res.text().unwrap();

    let mut file = File::create(path).unwrap();
    let _ = file.write_all(buffer.as_bytes());
    FetchResult {
        cached: false,
        html: buffer,
    }
}

fn main() {
    let cli = Cli::parse();
    // println!("Work Dir: {:?}", cli.work);
    // println!("Bible: {:?}", cli.bible);

    if !VERSION_URLS.contains_key(cli.bible.as_str()) {
        println!("Sorry, I don't know about Bible [{}].", cli.bible);
        println!("I do know about the following:");

        // Keys sorted in order.
        for (name, _) in VERSION_URLS.iter() {
            println!("  {}", name);
        }
        return;
    }

    match &cli.command {
        Some(Commands::Fetch { delay }) => {
            let client = reqwest::blocking::Client::builder()
                .user_agent(APP_USER_AGENT)
                .build()
                .unwrap();
            let mut url = VERSION_URLS[cli.bible.as_str()].to_string();
            println!("Fetch! [{}] with delay {} secs.", cli.bible, delay);
            let mut more = true;

            while more {
                let result = fetch_cache(&client, url.as_str());
                more = false;

                let document = scraper::Html::parse_document(&result.html);

                // TO FIX
                // We want to upgrade this to use CSS selectors.
                // For now, us the "working" code we have.
                let a_selector = scraper::Selector::parse("div>a").unwrap();
                for a in document.select(&a_selector) {
                    // Skip elements with a class attribute
                    if a.attr("class").is_some() {
                        continue;
                    }

                    if let Some(href) = a.attr("href") {
                        if href.contains("/bible/") {
                            let text = a.text().collect::<Vec<_>>();

                            if text.len() != 1 {
                                continue;
                            }
                            if text[0] != "Next Chapter" {
                                continue;
                            }

                            // Ok! We've found the Next Chapter a element!
                            if href.starts_with("/") {
                                url = String::from(BASE_URL) + href;
                            } else {
                                url = href.to_string();
                            }

                            // println!("Found HREF: {} => {}", href, url);
                            // panic!("Squirrel alert!");
                            more = true;
                            break;
                        }
                    }
                }

                if more {
                    if !result.cached {
                        thread::sleep(Duration::from_secs(*delay as u64));
                    }
                }
            }
            println!("I'm finished fetching!");
        }
        Some(Commands::Extract {}) => {
            println!("Extract...");
            let files = find_files(cli.work.to_str().unwrap(), cli.bible.as_str());
            for file in files.iter().take(5) {
                println!("File: {}", file);
                let buffer = std::fs::read_to_string(Path::new(file)).unwrap();
                let document = scraper::Html::parse_document(&buffer);

                let h1_selector = scraper::Selector::parse("h1").unwrap();
                let h1 = document.select(&h1_selector).next().unwrap();
                println!("h1 = {:?}", h1.text().collect::<Vec<_>>());

                let span_selector = scraper::Selector::parse("span").unwrap();
                for span in document.select(&span_selector) {
                    if let Some(data) = span.attr("data-usfm") {
                        println!("{}:", data);
                        let mut lines = Vec::<&str>::new();
                        for data_span in span.select(&span_selector) {
                            if let Some(data_class) = data_span.attr("class") {
                                if data_class.contains("content") {
                                    let mut text = data_span.text().collect::<Vec<_>>();
                                    println!("{} {:?}", data, text);
                                    lines.append(&mut text);
                                }
                            }
                        }
                        println!("{} {:?}", data, lines);
                    }
                }

                /* ESV 
                JHN.8.11 ["She said, “No one, Lord.” And Jesus said, "]
                JHN.8.11 ["“Neither do I condemn you; go, and from now on "]
                JHN.8.11 ["sin no more.”"]
                JHN.8.11 ["]]"]   <- What is this?
                JHN.8.11 ["  "]
                */
            }
        }
        Some(Commands::Test {}) => {
            println!("Testing...");
            // Test finding div>a[href^="/bible/"]
            let path = Path::new("bible").join("GEN.1.NIV");
            let buffer = std::fs::read_to_string(path).unwrap();
            let document = scraper::Html::parse_document(&buffer);

            // let a_selector = scraper::Selector::parse("div>a").unwrap();
            let a_selector = scraper::Selector::parse(r#"div>a[href ^="/bible/"]"#).unwrap();

            // This reduces down the number of items to check (from ~40 to 4)!
            // And invalidates the check for /bible/ !

            for a in document.select(&a_selector) {
                let text = a.text().collect::<Vec<_>>();
                println!("text: {:?}", text);
                if let Some(href) = a.attr("href") {
                    println!("href = {}", href);
                }
                println!("=====");
            }
        }
        None => {
            println!("Looking for FETCH or EXTRACT");
            println!("I've got nothing to do here...");
        }
    }

    /*
    return;


    let client = reqwest::blocking::Client::builder()
        .user_agent(APP_USER_AGENT)
        .build()
        .unwrap();

    if true {
        let mut url = String::from(URL);
        let mut previous_url = url.clone();
        let mut working = true;

        while working {
            // Begin the fetching process...
            let result = fetch_cache(&client, url.as_str());

            working = false;

            // Ok, HTML, get parsing!
            let document = scraper::Html::parse_document(&result.html);

            // For now, I don't care about parsing, just crawling/saving.
            // Locate the a href.
            let a_selector = scraper::Selector::parse("div>a").unwrap();
            for a in document.select(&a_selector) {
                let c = a.attr("class");
                if c.is_some() {
                    continue;
                }

                let href = a.attr("href");

                if href.is_some() {
                    let href = href.unwrap();
                    if href.contains("/bible/") {
                        let text = a.text().collect::<Vec<_>>();

                        if text.len() != 1 {
                            continue;
                        }
                        if text[0] != "Next Chapter" {
                            continue;
                        }

                        // previous_url = url;
                        url = href.to_string();
                        if url.starts_with("/") {
                            url = String::from("https://www.bible.com") + &url;
                        }

                        working = true;
                        break;
                    }
                }
            }
            if working {
                if !result.cached {
                    // Don't sleep if the results we just got were from the cache.
                    thread::sleep(Duration::from_secs(10));
                }
            }
        }
        println!("I'm finished.");
    } else {
        // let res = client.get("https://httpbin.org/anything").send().unwrap();
        // println!("anything: {}", res.text().unwrap());

        let mut file = File::open(Path::new("fetch1.html")).unwrap();
        let mut buffer = String::new();
        let _ = file.read_to_string(&mut buffer);
        drop(file);

        /*
        let res = client.get(URL).send().unwrap();
        let buffer = res.text().unwrap();
        println!("{}", res.text().unwrap());
        */

        let document = scraper::Html::parse_document(&buffer);
        let h1_selector = scraper::Selector::parse("h1").unwrap();

        let h1 = document.select(&h1_selector).next().unwrap();
        println!("h1 = {:?}", h1.text().collect::<Vec<_>>());

        // Selectors can't match on "if it exists" or
        // return a matching value, type fields.

        let span_selector = scraper::Selector::parse("span").unwrap();
        for span in document.select(&span_selector) {
            let data = span.attr("data-usfm");

            if data.is_some() {
                // Ok, iterate over these items...
                for data_span in span.select(&span_selector) {
                    let d = data_span.attr("class");
                    // println!("data_span: {:?}", data_span);
                    if d.is_some() {
                        let dt = d.unwrap();
                        if dt.contains("content") {
                            let text = data_span.text().collect::<Vec<_>>();
                            println!("{} {:?}", data.unwrap(), text)
                        }
                    }
                }

                /*
                GEN.1.25 ["And God made the beasts of the earth according to their kinds and the livestock according to their kinds, and everything that creeps on the ground according to its kind. And God saw that it was good."]
                GEN.1.25 ["  "]
                GEN.1.26 ["Then God said, "]
                GEN.1.26 ["“Let us make man"]
                GEN.1.26 [" in our image, "]
                GEN.1.26 ["after our likeness. And "]
                GEN.1.26 ["let them have dominion over the fish of the sea and over the birds of the heavens and over the livestock and over all the earth and over every creeping thing that creeps on the earth.”"]
                GEN.1.26 ["  "]
                GEN.1.27 ["So God created man in his own image,"]
                GEN.1.27 ["in the image of God he created him;"]
                GEN.1.27 ["  "]
                GEN.1.27 ["male and female he created them."]
                GEN.1.27 ["  "]
                 */

                /*
                let mut text = span.text().collect::<Vec<_>>();

                println!("span: {} {:?}", data.unwrap(), text);
                */
            }
        }
    }

    // let res = client.get("https://www.bible.com/bible/59/GEN.1.ESV").send().unwrap();
    */
}

/*
Book/chapter: <h1>
<h1>Genesis 1</h1>

<div data-usfm="">, <span data-usfm="">

<div data-usfm="GEN.1" class=
                        "ChapterContent_chapter__uvbXo">
                          <div class="ChapterContent_label__R2PLt">
                            1
                          </div>
                          <div class="ChapterContent_s1__bNNaW">
                            <span class=
                            "ChapterContent_heading__xBDcs">The
                            Creation of the World</span>
                          </div>
                          <div class="ChapterContent_p__dVKHb">
                            <span data-usfm="GEN.1.1" class=
                            "ChapterContent_verse__57FIw"><span class="ChapterContent_label__R2PLt">
                            1</span><span class=
                            "ChapterContent_content__RrUqA">In
                            the</span> <span class=
                            "ChapterContent_note__YlDW0 ChapterContent_x__tsTlk">
                            <span class=
                            "ChapterContent_label__R2PLt">#</span><span class="ChapterContent_body__O3qjr">Job
                            38:4-7; Ps. 33:6; 136:5; Isa. 42:5;
                            45:18; John 1:1-3; Acts 14:15; 17:24;
                            Col. 1:16, 17; Heb. 1:10; 11:3; Rev.
                            4:11</span></span><span class=
                            "ChapterContent_content__RrUqA">beginning,
                            God created the heavens and the
                            earth.</span></span>

   <span data-usfm=
                            "GEN.1.2" class=
                            "ChapterContent_verse__57FIw"><span class="ChapterContent_label__R2PLt">
                            2</span><span class=
                            "ChapterContent_content__RrUqA">The
                            earth was</span> <span class=
                            "ChapterContent_note__YlDW0 ChapterContent_x__tsTlk">
                            <span class=
                            "ChapterContent_label__R2PLt">#</span><span class="ChapterContent_body__O3qjr">Jer.
                            4:23</span></span><span class=
                            "ChapterContent_content__RrUqA">without
                            form and void, and darkness was over
                            the face of the deep. And the Spirit of
                            God was hovering over the face of the
                            waters.</span></span>
Next page link:

            <div class="[pointer-events:all]">
              <a href="/bible/59/GEN.2.ESV">


*/
